
__version__ = "6.10"
__author__ = "Alex Al-Saffar"
__date__ = "2022-03-16"
docs = r"https://crocodile.readthedocs.io/en/latest/"
git = r"https://github.com/thisismygitrepo/crocodile"
release_notes = """
major bux fixes.
"""
